import unittest
import numpy as np
from .bvgamma import *
from flavio.physics.bdecays.formfactors.b_v import bsz_parameters
from flavio.physics.eft import WilsonCoefficients
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.parameters import default_parameters
import flavio


wc = WilsonCoefficients()
par = default_parameters#

class TestBVgamma(unittest.TestCase):
    def test_bksgamma(self):
        # just check if this works
        flavio.Observable["ACP(B0->K*gamma)"].prediction_central(par, wc)
        flavio.Observable["S_K*gamma"].prediction_central(par, wc)
        # numerical comparison to  David's old Mathematica code
        self.assertAlmostEqual(
            flavio.Observable["BR(B0->K*gamma)"].prediction_central(par, wc)*1e5/3.91526,
             1, places=1)
        self.assertAlmostEqual(
            flavio.Observable["BR(B+->K*gamma)"].prediction_central(par, wc)*1e5/4.11625,
             1, places=1)

    def test_bksgamma(self):
        # just check if this works
        flavio.Observable["ACP(Bs->phigamma)"].prediction_central(par, wc)
        flavio.Observable["S_phigamma"].prediction_central(par, wc)
        flavio.Observable["BR(Bs->phigamma)"].prediction_central(par, wc)
        flavio.Observable["ADeltaGamma(Bs->phigamma)"].prediction_central(par, wc)



"""Functions for parsing the parameter data files"""

import yaml
import pkgutil
from flavio.classes import *
from flavio.statistics.probability import *
from flavio._parse_errors import errors_from_string
import flavio
import re
from flavio.measurements import _fix_correlation_matrix
from math import sqrt


def _read_yaml_object_metadata(obj, constraints):
    parameters = yaml.load(obj)
    for parameter_name, info in parameters.items():
        p = Parameter(parameter_name)
        if 'description' in info and info['description'] is not None:
            p.description = info['description']
        if 'tex' in info and info['tex'] is not None:
            p.tex = info['tex']

def read_file_metadata(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_metadata(f, constraints)

def _read_yaml_object_values(obj, constraints):
    parameters = yaml.load(obj)
    for parameter_name, value in parameters.items():
        p = Parameter[parameter_name] # this will raise an error if the parameter doesn't exist!
        constraints.set_constraint(parameter_name, value)

def _read_yaml_object_new(obj):
    """Read parameter constraints from a YAML stream or file that are compatible
    with the format generated by the `get_yaml` method of
    `flavio.classes.ParameterConstraints`."""
    parameters = yaml.load(obj)
    return ParameterConstraints.from_yaml_dict(parameters)

def _read_yaml_object_values_correlated(obj, constraints):
    list_ = yaml.load(obj)
    for parameter_group in list_:
        parameter_names = []
        central_values = []
        errors = []
        for dict_list in parameter_group['values']:
            parameter_name, value = list(dict_list.items())[0]
            Parameter[parameter_name] # this will raise an error if the parameter doesn't exist!
            parameter_names.append(parameter_name)
            error_dict = errors_from_string(value)
            central_values.append(error_dict['central_value'])
            squared_error = 0.
            for sym_err in error_dict['symmetric_errors']:
                squared_error += sym_err**2
            for asym_err in error_dict['asymmetric_errors']:
                squared_error += asym_err[0]*asym_err[1]
            errors.append(sqrt(squared_error))
        correlation = _fix_correlation_matrix(parameter_group['correlation'], len(parameter_names))
        covariance = np.outer(np.asarray(errors), np.asarray(errors))*correlation
        if not np.all(np.linalg.eigvals(covariance) > 0):
            # if the covariance matrix is not positive definite, try a dirty trick:
            # multiply all the correlations by 0.99.
            n_dim = len(correlation)
            correlation = (correlation - np.eye(n_dim))*0.99 + np.eye(n_dim)
            covariance = np.outer(np.asarray(errors), np.asarray(errors))*correlation
            # if it still isn't positive definite, give up.
            assert np.all(np.linalg.eigvals(covariance) > 0), "The covariance matrix is not positive definite!" + str(covariance)
        constraints.add_constraint(parameter_names, MultivariateNormalDistribution(central_values, covariance))

def read_file(filename):
    """Read parameter values from a YAML file in the format generated by the
    `get_yaml` method of the `ParameterConstraints` class, returning a
    `ParameterConstraints` instance."""
    with open(filename, 'r') as f:
        return _read_yaml_object_new(f)

def read_file_values(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_values(f, constraints)

def read_file_values_correlated(filename, constraints):
    """Read parameter values from a YAML file."""
    with open(filename, 'r') as f:
        _read_yaml_object_values_correlated(f, constraints)

def write_file(filename, constraints):
    """Write parameter constraints to a YAML file."""
    with open(filename, 'w') as f:
        yaml.dump(constraints.get_yaml_dict(), f)


# particles from the PDG data file whose mass we're interested in)
pdg_include = ['B(s)', 'B(c)', 'B(s)*', 'B*+', 'B*0', 'B+', 'B0', 'D(s)', 'D(s)*', 'D+', 'D0',
               'H', 'J/psi(1S)', 'K(L)', 'K(S)', 'K*(892)+', 'K*(892)0', 'K+', 'K0',
               'Lambda', 'Lambda(b)', 'Lambda(c)', 'omega(782)', 'D*(2007)', 'D*(2010)',
               'W', 'Z',  'b',  'c', 'd', 'e', 'eta', 'f(0)(980)',
               'mu',  'phi(1020)', 'pi+', 'pi0', 'psi(2S)', 'rho(770)+', 'rho(770)0',
               's', 't', 'tau', 'u']
# dictionary translating PDG particle names to the ones in the code.
pdg_translate = {
    'B(s)': 'Bs',
    'B(c)': 'Bc',
    'D(s)': 'Ds',
    'B(s)*': 'Bs*',
    'D(s)*': 'Ds*',
    'D*(2007)' : 'D*0',
    'D*(2010)' : 'D*+',
    'J/psi(1S)': 'J/psi',
    'K(L)': 'KL',
    'K(S)': 'KS',
    'K*(892)+': 'K*+',
    'K*(892)0': 'K*0',
    'phi(1020)': 'phi',
    'rho(770)0': 'rho0',
    'rho(770)+': 'rho+',
    'f(0)(980)': 'f0',
    "eta'(958)": "eta'",
    'omega(782)': 'omega',
    'Lambda(b)' : 'Lambdab',
    'Lambda(c)' : 'Lambdac',
    'Higgs' : 'h', # this is necessary for the 2013 data file
    'H' : 'h',
    }

def _read_pdg_masswidth(filename):
    """Read the PDG mass and width table and return a dictionary.

    Parameters
    ----------
    filname : string
        Path to the PDG data file, e.g. 'data/pdg/mass_width_2015.mcd'

    Returns
    -------
    particles : dict
        A dictionary where the keys are the particle names with the charge
        appended in case of a multiplet with different masses, e.g. 't'
        for the top quark, 'K+' and 'K0' for kaons.
        The value of the dictionary is again a dictionary with the following
        keys:
        - 'id': PDG particle ID
        - 'mass': list with the mass, postitive and negative error in GeV
        - 'width': list with the width, postitive and negative error in GeV
        - 'name': same as the key
    """
    data = pkgutil.get_data('flavio.physics', filename)
    lines = data.decode('utf-8').splitlines()
    particles_by_name = {}
    for line in lines:
        if  line.strip()[0] == '*':
            continue
        mass = ((line[33:51]), (line[52:60]), (line[61:69]))
        mass = [float(m) for m in mass]
        width = ((line[70:88]), (line[89:97]), (line[98:106]))
        if  width[0].strip() == '':
            width = (0,0,0)
        else:
            width = [float(w) for w in width]
        ids = line[0:32].split()
        charges = line[107:128].split()[1].split(',')
        if len(ids) != len(charges):
            raise ValueError()
        for i, id_ in enumerate(ids):
            particle = {}
            particle_charge = charges[i].strip()
            particle[particle_charge] = {}
            particle[particle_charge]['id'] = id_.strip()
            particle[particle_charge]['mass']  = mass
            particle[particle_charge]['charge']  = particle_charge
            particle[particle_charge]['width'] = width
            particle_name = line[107:128].split()[0]
            particle[particle_charge]['name'] = particle_name
            if particle_name in particles_by_name.keys():
                particles_by_name[particle_name].update(particle)
            else:
                particles_by_name[particle_name] = particle
    result = { k + kk: vv for k, v in particles_by_name.items() for kk, vv in v.items() if len(v) > 1}
    result.update({ k: list(v.values())[0] for k, v in particles_by_name.items() if len(v) == 1})
    return result

def _pdg_particle_string_to_tex(string):
    regex = re.compile(r"^([A-Za-z]+)([\*+-0]*)(?:\((\d*[A-Za-z]+)\))*(?:\((\d+)\))*([\*+-0]*)$")
    m = regex.match(string)
    if m is None:
        if string=='J/psi(1S)':
            return r'J/\psi'
        return string
    sup = m.group(2) + m.group(5)
    sub = m.group(3)
    out = m.group(1)
    if len(out) > 1:
        out = "\\" + out
    if sup is not None and sup != '':
        out += r'^{' + sup + r'}'
    if sub is not None and sub != '':
        out += r'_{' + sub + r'}'
    return out

def read_pdg(year, constraints):
    """Read particle masses and widths from the PDG data file of a given year."""
    particles = _read_pdg_masswidth('data/pdg/mass_width_' + str(year) + '.mcd')
    for particle in pdg_include:
        parameter_name = 'm_' + pdg_translate.get(particle, particle) # translate if necessary
        tex_name = _pdg_particle_string_to_tex(particle)
        try:
            # if parameter already exists, remove existing constraints on it
            m = Parameter[parameter_name]
            constraints.remove_constraint(parameter_name)
        except KeyError:
            # otherwise, create it
            m = Parameter(parameter_name)
        m.tex = r'$m_{' + tex_name + '}$'
        if tex_name =='b' or tex_name == 'c':
            m.tex = r'$m_{' + tex_name + '}(m_{' + tex_name + '})$'
            m.description = r'$' + tex_name + r'$ quark mass in the $\overline{\text{MS}}$ scheme at the scale $m_' + tex_name + r'$'
        elif tex_name =='s' or tex_name =='u' or tex_name =='d':
            m.tex = r'$m_{' + tex_name + r'}(2\,\text{GeV})$'
            m.description = r'$' + tex_name + r'$ quark mass in the $\overline{\text{MS}}$ scheme at 2 GeV'
        elif tex_name =='t':
            m.description = r'$' + tex_name + r'$ quark pole mass'
        else:
            m.description = r'$' + tex_name + r'$ mass'
        m_central, m_right, m_left = particles[particle]['mass']
        m_left = abs(m_left) # make left error positive
        if m_right == m_left:
            constraints.add_constraint([parameter_name],
                NormalDistribution(m_central, m_right))
        else:
            constraints.add_constraint([parameter_name],
                AsymmetricNormalDistribution(m_central,
                right_deviation=m_right, left_deviation=m_left))
        if particles[particle]['width'][0] == 0: # 0 is for particles where the width is unknown (e.g. B*)
            continue
        G_central, G_right, G_left = particles[particle]['width']
        G_left = abs(G_left) # make left error positive
        parameter_name = 'tau_' + pdg_translate.get(particle, particle) # translate if necessary
        try:
            # if parameter already exists, remove existing constraints on it
            tau = Parameter[parameter_name]
            constraints.remove_constraint(parameter_name)
        except KeyError:
            # otherwise, create it
            tau = Parameter(parameter_name)
        tau.tex = r'$\tau_{' + tex_name + '}$'
        tau.description = r'$' + tex_name + r'$ lifetime'
        tau_central = 1/G_central # life time = 1/width
        tau_left = G_left/G_central**2
        tau_right = G_right/G_central**2
        if tau_left == tau_right:
            constraints.add_constraint([parameter_name],
                NormalDistribution(tau_central, tau_right))
        else:
            constraints.add_constraint([parameter_name],
                AsymmetricNormalDistribution(tau_central,
                    right_deviation=tau_right, left_deviation=tau_left))



############### Read default parameters ###################

# Create the object
default_parameters = ParameterConstraints()

# Read the parameter metadata from the default YAML data file
_read_yaml_object_metadata(pkgutil.get_data('flavio', 'data/parameters_metadata.yml'), default_parameters)

# Read the uncorrelated parameter values from the default YAML data file
_read_yaml_object_values(pkgutil.get_data('flavio', 'data/parameters_uncorrelated.yml'), default_parameters)

# Read the correlated parameter values from the default YAML data file
_read_yaml_object_values_correlated(pkgutil.get_data('flavio', 'data/parameters_correlated.yml'), default_parameters)

# Read the parameters from the default PDG data file
read_pdg(2017, default_parameters)

# Read default parameters for B->V form factors
## first load LCSR-only form factors
flavio.physics.bdecays.formfactors.b_v.bsz_parameters.bsz_load_v2_lcsr(default_parameters)
## then load combined LCSR-lattice fits. Overwrites LCSR ones for B->K*, Bs->K*, Bs->phi, but not B->rho, B->omega
flavio.physics.bdecays.formfactors.b_v.bsz_parameters.bsz_load_v2_combined(default_parameters)

# Read default parameters for Lambdab->Lambda form factors
flavio.physics.bdecays.formfactors.lambdab_12.lattice_parameters.lattice_load_ho(default_parameters)



"""Functions for reading a set of Wilson coefficients from an FLHA file.
This assumes the normalization convention used by SPheno modules generated
by FlavorKit with SARAH 4.8.1+ and should not be used with other FLHA-compatible
files, as the normalization is not fixed by the accord."""


import flavio.io.slha
import flavio
from math import pi, sqrt
import os
import logging

log = logging.getLogger('SLHA')
log.setLevel('WARNING')

def _prefactors_bsll(par, scale):
    GF = par['GF']
    alpha_e = flavio.physics.running.running.get_alpha(par, scale)['alpha_e']
    xi_t = flavio.physics.ckm.xi('t', 'bs')(par)
    pre_all = -4*GF/sqrt(2)*xi_t
    pre_910 = pre_all * alpha_e/(4*pi)
    pre_7 =   pre_all /(16*pi**2)
    pre_8 =   pre_all /(16*pi**2)
    pre_nu =  pre_all * alpha_e/(4*pi)
    return {
            'C7_bs': pre_7,
            'C7p_bs': pre_7,
            'C8_bs': pre_8,
            'C8p_bs': pre_8,
            'C9_bsee': pre_910,
            'C9p_bsee': pre_910,
            'C10_bsee': pre_910,
            'C10p_bsee': pre_910,
            'C9_bsmumu': pre_910,
            'C9p_bsmumu': pre_910,
            'C10_bsmumu': pre_910,
            'C10p_bsmumu': pre_910,
            'CL_bsnuenue': pre_nu,
            'CR_bsnuenue': pre_nu,
            'CL_bsnumunumu': pre_nu,
            'CR_bsnumunumu': pre_nu,
            'CL_bsnutaunutau': pre_nu,
            'CR_bsnutaunutau': pre_nu,
           }


_flha_dict ={
(305, 4422): 'C7_bs',
(305, 4322): 'C7p_bs',
(305, 6421): 'C8_bs',
(305, 6321): 'C8p_bs',
(3051111, 4133): 'C9_bsee',
(3051111, 4233): 'C9p_bsee',
(3051111, 4137): 'C10_bsee',
(3051111, 4237): 'C10p_bsee',
(3051313, 4133): 'C9_bsmumu',
(3051313, 4233): 'C9p_bsmumu',
(3051313, 4137): 'C10_bsmumu',
(3051313, 4237): 'C10p_bsmumu',
(3051212, 4141): 'CL_bsnuenue',
(3051212, 4241): 'CR_bsnuenue',
(3051414, 4141): 'CL_bsnumunumu',
(3051414, 4241): 'CR_bsnumunumu',
(3051616, 4141): 'CL_bsnutaunutau',
(3051616, 4241): 'CR_bsnutaunutau',
}

def read_wilson(filename):
    r"""Read new physics contributions to Wilson coefficients from an output file
    in FLHA format produced by a SPheno module generated with SARAH 4.9.0+
    with FlavorKit.

    *Caution*: this function should not be used with FLHA files produced by any
    other code, in particular not the default version of SPheno or older
    SARAH/FlavorKit versions, as they use different normalization conventions
    for the Wilson coefficients.

    Input
    -----

    - `filename`: the path to the file as a string

    Returns an instance of `flavio.WilsonCoefficients` that can be used for
    `flavio.np_prediction`, for instance.
    """
    if not os.path.exists(filename):
        raise ValueError("File " + filename + " not found.")
    card = flavio.io.slha.read(filename)
    wc_flha = card.matrices['fwcoef']
    scale = wc_flha.scale
    par_dict = flavio.default_parameters.get_central_all()
    prefac = _prefactors_bsll(par_dict, scale)
    wc_dict = {}
    for k, v in wc_flha.dict().items():
        if k[-1] != 1: # only look at NP-only Wilson coefficients
            continue
        if k[:2] not in _flha_dict:
            log.warning('Wilson coefficient ' + str(k[:2]) + ' unknown to flavio; ignored.')
            continue
        wc_name = _flha_dict[k[:2]]
        if wc_name not in prefac:
            wc_dict[wc_name] = v
        else:
            wc_dict[wc_name] = v / prefac[wc_name]
    wc_obj =  flavio.WilsonCoefficients()
    wc_obj.set_initial(wc_dict, scale=scale)
    return wc_obj

def read_ckm(filename, par_constraints):
    r"""Read CKM Wolfenstein parameters from the `VCKMIN` block of an FLHA
    file.

    Input
    -----

    - `filename`: the path to the file as a string
    - `par_constraints`: an instance of `flavio.ParameterConstraints`, e.g.
    `flavio.default_parameters`

    Note that you have to set the config option
    `config['implementation']['CKM matrix']`
    to `Wolfenstein` first; otherwise a warning is issued.

    *Caution*: since the FLHA/SLHA format does not specify uncertainties,
    the parameters are assumed to be known exactly.
    """
    if not os.path.exists(filename):
        raise ValueError("File " + filename + " not found.")
    if flavio.config['implementation']['CKM matrix'] != 'Wolfenstein':
        log.warning('CKM matrix parametrization is not set to "Wolfenstein". read_ckm will have no effect!')
    card = flavio.io.slha.read(filename)
    try:
        wc_flha = card.blocks['vckmin']
        try:
            par_constraints.set_constraint('laC', wc_flha[1])
            par_constraints.set_constraint('A', wc_flha[2])
            par_constraints.set_constraint('rhobar', wc_flha[3])
            par_constraints.set_constraint('etabar', wc_flha[4])
        except KeyError:
            raise KeyError("One of the Wolfenstein parameters seems to be missing from the VCKMIN block")
    except KeyError:
        raise ValueError("This file does not contain a VCKMIN block")



"""Functions for exclusive $B\to V\ell^+\ell^-$ decays."""

from math import sqrt, pi, log
import numpy as np
from flavio.physics.bdecays.common import lambda_K, beta_l, meson_quark, meson_ff
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.physics.bdecays import matrixelements
from flavio.physics import ckm
from flavio.config import config
from flavio.physics.running import running
from .amplitudes import *
from scipy.integrate import quad
from flavio.classes import Observable, Prediction
import flavio

def dGdq2(J):
    return 3/4. * (2 * J['1s'] + J['1c']) - 1/4. * (2 * J['2s'] + J['2c'])

def dGdq2_ave(J, J_bar):
    return ( dGdq2(J) + dGdq2(J_bar) )/2.

def dGdq2_diff(J, J_bar):
    return ( dGdq2(J) - dGdq2(J_bar) )/2.

# denominator of S_i and A_i observables
def SA_den(J, J_bar):
    return 2*dGdq2_ave(J, J_bar)

# denominator of P_i observables
def P_den(J, J_bar):
    return S_theory_num(J, J_bar, '2s')

def S_theory(J, J_bar, i):
    r"""CP-averaged angular observable $S_i$ in the theory convention."""
    return S_theory_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def S_theory_num(J, J_bar, i):
    return (J[i] + J_bar[i])

def A_theory(J, J_bar, i):
    r"""Angular CP asymmetry $A_i$ in the theory convention."""
    return A_theory_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def A_theory_num(J, J_bar, i):
    return (J[i] - J_bar[i])

def S_experiment(J, J_bar, i):
    r"""CP-averaged angular observable $S_i$ in the LHCb convention.

    See eq. (C.8) of arXiv:1506.03970v2.
    """
    return S_experiment_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def S_experiment_num(J, J_bar, i):
    if i in [4, '6s', '6c', 7, 9]:
        return -S_theory_num(J, J_bar, i)
    return S_theory_num(J, J_bar, i)

def A_experiment(J, J_bar, i):
    r"""Angular CP asymmetry $A_i$ in the LHCb convention.

    See eq. (C.8) of arXiv:1506.03970v2.
    """
    return A_experiment_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def A_experiment_num(J, J_bar, i):
    if i in [4, '6s', '6c', 7, 9]:
        return -A_theory_num(J, J_bar, i)
    return A_theory_num(J, J_bar, i)

def AFB_experiment(J, J_bar):
    r"""Forward-backward asymmetry in the LHCb convention.

    See eq. (C.9) of arXiv:1506.03970v2.
    """
    return AFB_experiment_num(J, J_bar)/SA_den(J, J_bar)

def AFB_experiment_num(J, J_bar):
    return 3/4.*S_experiment_num(J, J_bar, '6s')

def AFB_theory(J, J_bar):
    """Forward-backward asymmetry in the original theory convention.
    """
    return AFB_theory_num(J, J_bar)/SA_den(J, J_bar)

def AFB_theory_num(J, J_bar):
    return 3/4.*S_theory_num(J, J_bar, '6s')

def FL(J, J_bar):
    r"""Longitudinal polarization fraction $F_L$"""
    return FL_num(J, J_bar)/SA_den(J, J_bar)

def FL_num(J, J_bar):
    return -S_theory_num(J, J_bar, '2c')

def FLhat(J, J_bar):
    r"""Modified longitudinal polarization fraction for vanishing lepton masses,
    $\hat F_L$.

    See eq. (32) of arXiv:1510.04239.
    """
    return FLhat_num(J, J_bar)/SA_den(J, J_bar)

def FLhat_num(J, J_bar):
    return -S_theory_num(J, J_bar, '1c')


def bvll_obs(function, q2, wc_obj, par, B, V, lep):
    ml = par['m_'+lep]
    mB = par['m_'+B]
    mV = par['m_'+V]
    if q2 < 4*ml**2 or q2 > (mB-mV)**2:
        return 0
    scale = config['renormalization scale']['bvll']
    mb = running.get_mb(par, scale)
    ff = get_ff(q2, par, B, V)
    h = helicity_amps(q2, wc_obj, par, B, V, lep)
    h_bar = helicity_amps_bar(q2, wc_obj, par, B, V, lep)
    J = angular.angularcoeffs_general_v(h, q2, mB, mV, mb, 0, ml, ml)
    J_bar = angular.angularcoeffs_general_v(h_bar, q2, mB, mV, mb, 0, ml, ml)
    return function(J, J_bar)

def bvll_dbrdq2(q2, wc_obj, par, B, V, lep):
    tauB = par['tau_'+B]
    return tauB * bvll_obs(dGdq2_ave, q2, wc_obj, par, B, V, lep)

def nintegrate_pole(function, q2min, q2max, epsrel=0.005):
    # this is a special integration function to treat the presence of the
    # photon pole at low q^2. If q2min is below 0.1 GeV^2, it adds and subtracts
    # the 1/q^2-enhanced pole part to split the integral into a well-behaved part
    # and one that is trivially solved analytically.
    # This leads to a huge speed-up.
    if q2min <= 0.1 and q2min > 0:
        q20 = q2min
        f_q20 = function(q20)
        int_a = flavio.math.integrate.nintegrate(lambda q2: function(q2)-f_q20*q20/q2, q2min, q2max, epsrel=epsrel)
        int_b = f_q20*q20 * log(q2max/q2min)
        return int_a + int_b
    else:
        return flavio.math.integrate.nintegrate(function, q2min, q2max)

def bvll_obs_int(function, q2min, q2max, wc_obj, par, B, V, lep, epsrel=0.005):
    def obs(q2):
        return bvll_obs(function, q2, wc_obj, par, B, V, lep)
    _q2min = max(q2min, 4*par['m_'+lep]**2) # clip to physical range
    return nintegrate_pole(obs, _q2min, q2max, epsrel=epsrel)

def bvll_dbrdq2_int(q2min, q2max, wc_obj, par, B, V, lep, epsrel=0.005):
    def obs(q2):
        return bvll_dbrdq2(q2, wc_obj, par, B, V, lep)
    _q2min = max(q2min, 4*par['m_'+lep]**2) # clip to physical range
    return nintegrate_pole(obs, _q2min, q2max, epsrel=epsrel)/(q2max-_q2min)

# Functions returning functions needed for Prediction instances

def bvll_dbrdq2_int_func(B, V, lep):
    def fct(wc_obj, par, q2min, q2max):
        return bvll_dbrdq2_int(q2min, q2max, wc_obj, par, B, V, lep)
    return fct

def bvll_dbrdq2_func(B, V, lep):
    def fct(wc_obj, par, q2):
        return bvll_dbrdq2(q2, wc_obj, par, B, V, lep)
    return fct

def bvll_obs_int_ratio_func(func_num, func_den, B, V, lep):
    def fct(wc_obj, par, q2min, q2max):
        num = bvll_obs_int(func_num, q2min, q2max, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom = bvll_obs_int(func_den, q2min, q2max, wc_obj, par, B, V, lep)
        return num/denom
    return fct

def bvll_obs_int_ratio_leptonflavour(func, B, V, l1, l2):
    def fct(wc_obj, par, q2min, q2max):
        num = bvll_obs_int(func, q2min, q2max, wc_obj, par, B, V, l1, epsrel=0.0005)
        if num == 0:
            return 0
        denom = bvll_obs_int(func, q2min, q2max, wc_obj, par, B, V, l2, epsrel=0.0005)
        return num/denom
    return fct

def bvll_obs_ratio_leptonflavour(func, B, V, l1, l2):
    def fct(wc_obj, par, q2):
        num = bvll_obs(func, q2, wc_obj, par, B, V, l1)
        if num == 0:
            return 0
        denom = bvll_obs(func, q2, wc_obj, par, B, V, l2)
        return num/denom
    return fct

def bvll_obs_ratio_func(func_num, func_den, B, V, lep):
    def fct(wc_obj, par, q2):
        num = bvll_obs(func_num, q2, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom = bvll_obs(func_den, q2, wc_obj, par, B, V, lep)
        return num/denom
    return fct

# function needed for the P' "optimized" observables
# note that this is the convention used by LHCb, NOT the one used in 1303.5794!
def bvll_pprime_func(func_num, B, V, lep):
    def fct(wc_obj, par, q2):
        num = bvll_obs(func_num, q2, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom_2s = bvll_obs(lambda J, J_bar: S_experiment_num(J, J_bar, '2s'), q2, wc_obj, par, B, V, lep)
        denom_2c = bvll_obs(lambda J, J_bar: S_experiment_num(J, J_bar, '2c'), q2, wc_obj, par, B, V, lep)
        denom = 2*sqrt(-denom_2s*denom_2c)
        return num/denom
    return fct

def bvll_pprime_int_func(func_num, B, V, lep):
    def fct(wc_obj, par, q2min, q2max):
        num = bvll_obs_int(func_num, q2min, q2max, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom_2s = bvll_obs_int(lambda J, J_bar: S_experiment_num(J, J_bar, '2s'), q2min, q2max, wc_obj, par, B, V, lep)
        denom_2c = bvll_obs_int(lambda J, J_bar: S_experiment_num(J, J_bar, '2c'), q2min, q2max, wc_obj, par, B, V, lep)
        denom = 2*sqrt(-denom_2s*denom_2c)
        return num/denom
    return fct


# Observable and Prediction instances

_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_observables = {
'ACP': {'func_num': dGdq2_diff, 'tex': r'A_\text{CP}', 'desc': 'Direct CP asymmetry'},
'AFB': {'func_num': AFB_experiment_num, 'tex': r'A_\text{FB}', 'desc': 'forward-backward asymmetry'},
'FL': {'func_num': FL_num, 'tex': r'F_L', 'desc': 'longitudinal polarization fraction'},
'S3': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 3), 'tex': r'S_3', 'desc': 'CP-averaged angular observable'},
'S4': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 4), 'tex': r'S_4', 'desc': 'CP-averaged angular observable'},
'S5': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 5), 'tex': r'S_5', 'desc': 'CP-averaged angular observable'},
'S6c': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, '6c'), 'tex': r'S_6^c', 'desc': 'CP-averaged angular observable'},
'S7': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 7), 'tex': r'S_7', 'desc': 'CP-averaged angular observable'},
'S8': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 8), 'tex': r'S_8', 'desc': 'CP-averaged angular observable'},
'S9': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 9), 'tex': r'S_9', 'desc': 'CP-averaged angular observable'},
'A3': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 3), 'tex': r'A_3', 'desc': 'Angular CP asymmetry'},
'A4': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 4), 'tex': r'A_4', 'desc': 'Angular CP asymmetry'},
'A5': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 5), 'tex': r'A_5', 'desc': 'Angular CP asymmetry'},
'A6s': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, '6s'), 'tex': r'A_6^s', 'desc': 'Angular CP asymmetry'},
'A7': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 7), 'tex': r'A_7', 'desc': 'Angular CP asymmetry'},
'A8': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 8), 'tex': r'A_8', 'desc': 'Angular CP asymmetry'},
'A9': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 9), 'tex': r'A_9', 'desc': 'Angular CP asymmetry'},
}
# for the P observables, the convention of LHCb is used. This differs by a
# sign in P_2 and P_3 from the convention in arXiv:1303.5794
_observables_p = {
'P1': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 3)/2., 'tex': r'P_1', 'desc': "CP-averaged \"optimized\" angular observable"},
'P2': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, '6s')/8., 'tex': r'P_2', 'desc': "CP-averaged \"optimized\" angular observable"},
'P3': {'func_num': lambda J, J_bar: -S_experiment_num(J, J_bar, 9)/4., 'tex': r'P_3', 'desc': "CP-averaged \"optimized\" angular observable"},
'ATIm': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 9)/2., 'tex': r'A_T^\text{Im}', 'desc': "Transverse CP asymmetry"},
}
_observables_pprime = {
'P4p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 4), 'tex': r'P_4^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
'P5p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 5), 'tex': r'P_5^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
# yes, P6p depends on J_7, not J_6. Don't ask why.
'P6p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 7), 'tex': r'P_6^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
'P8p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 8), 'tex': r'P_8^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
}
_hadr = {
'B0->K*': {'tex': r"B^0\to K^{\ast 0}", 'B': 'B0', 'V': 'K*0', },
'B+->K*': {'tex': r"B^+\to K^{\ast +}", 'B': 'B+', 'V': 'K*+', },
}

for l in ['e', 'mu', 'tau']:
    for M in _hadr.keys():

        _process_tex = _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+r"^-"
        _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _process_tex + r"$"

        for obs in sorted(_observables.keys()):

            # binned angular observables
            _obs_name = "<" + obs + ">("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description('Binned ' + _observables[obs]['desc'] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$\langle " + _observables[obs]['tex'] + r"\rangle(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_obs_int_ratio_func(_observables[obs]['func_num'], SA_den, _hadr[M]['B'], _hadr[M]['V'], l))

            # differential angular observables
            _obs_name = obs + "("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2'])
            _obs.set_description(_observables[obs]['desc'][0].capitalize() + _observables[obs]['desc'][1:] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$" + _observables[obs]['tex'] + r"(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_obs_ratio_func(_observables[obs]['func_num'], SA_den, _hadr[M]['B'], _hadr[M]['V'], l))

        for obs in sorted(_observables_p.keys()):

            # binned "optimized" angular observables P
            _obs_name = "<" + obs + ">("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description('Binned ' + _observables_p[obs]['desc'] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$\langle " + _observables_p[obs]['tex'] + r"\rangle(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_obs_int_ratio_func(_observables_p[obs]['func_num'], P_den, _hadr[M]['B'], _hadr[M]['V'], l))

            # differential "optimized"  angular observables
            _obs_name = obs + "("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2'])
            _obs.set_description(_observables_p[obs]['desc'][0].capitalize() + _observables_p[obs]['desc'][1:] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$" + _observables_p[obs]['tex'] + r"(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_obs_ratio_func(_observables_p[obs]['func_num'], P_den, _hadr[M]['B'], _hadr[M]['V'], l))

        for obs in sorted(_observables_pprime.keys()):

            # binned "optimized" angular observables P'
            _obs_name = "<" + obs + ">("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description('Binned ' + _observables_pprime[obs]['desc'] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$\langle " + _observables_pprime[obs]['tex'] + r"\rangle(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_pprime_int_func(_observables_pprime[obs]['func_num'], _hadr[M]['B'], _hadr[M]['V'], l))

            # differential "optimized"  angular observables
            _obs_name = obs + "("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2'])
            _obs.set_description(_observables_pprime[obs]['desc'][0].capitalize() + _observables_pprime[obs]['desc'][1:] + r" in $" + _process_tex + r"$")
            _obs.tex = r"$" + _observables_pprime[obs]['tex'] + r"(" + _process_tex + r")$"
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, bvll_pprime_func(_observables_pprime[obs]['func_num'], _hadr[M]['B'], _hadr[M]['V'], l))

        # binned branching ratio
        _obs_name = "<dBR/dq2>("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description(r"Binned differential branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\langle \frac{d\text{BR}}{dq^2} \rangle(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        Prediction(_obs_name, bvll_dbrdq2_int_func(_hadr[M]['B'], _hadr[M]['V'], l))

        # differential branching ratio
        _obs_name = "dBR/dq2("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(r"Differential branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\frac{d\text{BR}}{dq^2}(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        Prediction(_obs_name, bvll_dbrdq2_func(_hadr[M]['B'], _hadr[M]['V'], l))

# Lepton flavour ratios
for l in [('mu','e'), ('tau','mu'),]:
    for M in _hadr.keys():

        # binned ratio of BRs
        _obs_name = "<R"+l[0]+l[1]+">("+M+"ll)"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description(r"Ratio of partial branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
        _obs.tex = r"$\langle R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} \rangle(" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
        for li in l:
            # add taxonomy for both processes (e.g. B->Vee and B->Vmumu)
            _obs.add_taxonomy(r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _hadr[M]['tex'] +_tex[li]+r"^+"+_tex[li]+r"^-$")
        Prediction(_obs_name, bvll_obs_int_ratio_leptonflavour(dGdq2_ave, _hadr[M]['B'], _hadr[M]['V'], *l))

        # differential ratio of BRs
        _obs_name = "R"+l[0]+l[1]+"("+M+"ll)"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(r"Ratio of differential branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
        _obs.tex = r"$R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} (" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
        for li in l:
            # add taxonomy for both processes (e.g. B->Vee and B->Vmumu)
            _obs.add_taxonomy(r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _hadr[M]['tex'] +_tex[li]+r"^+"+_tex[li]+r"^-$")
        Prediction(_obs_name, bvll_obs_ratio_leptonflavour(dGdq2_ave, _hadr[M]['B'], _hadr[M]['V'], *l))



r"""Functions for $B\to\ell\nu$ decays."""

import flavio
from math import pi
from flavio.physics.bdecays.common import meson_quark

def br_plnu_general(wc, par, Vij, P, qiqj, lep, nu, mq1, mq2, delta=0):
    r"""Branching ratio of general $P^+\to\ell^+\nu_\ell$ decay.

    `Vij` is the appropriate CKM matrix element.
    `mq1` and `mq2` are the masses of the two quarks forming the meson $P$.
    `delta` (detaults to 0) is a correction factor to account for different
    experimental treatment of electromagnetic effects, for instance.
    """
    ml = par['m_'+lep]
    mP = par['m_'+P]
    GF = par['GF']
    tau = par['tau_'+P]
    f = par['f_'+P]
    # Wilson coefficient dependence
    qqlnu = qiqj + lep + 'nu' + nu
    rWC = (wc['CVL_'+qqlnu] - wc['CVR_'+qqlnu]) + mP**2/ml/(mq1 + mq2) * (wc['CSR_'+qqlnu] - wc['CSL_'+qqlnu])
    N = tau * GF**2 * f**2 / (8*pi) * mP * ml**2  * (1 - ml**2/mP**2)**2
    return N * abs(Vij)**2 * abs(rWC)**2 * (1 + delta)


def br_blnu(wc_obj, par, B, lep):
    return sum([_br_blnu(wc_obj,par,B,lep,nu) for nu in ['e', 'mu', 'tau']])


def _br_blnu(wc_obj, par, B, lep, nu):
    r"""Branching ratio of $B_q\to\ell^+\nu_\ell$."""
    bq = meson_quark[B]
    # CKM element
    if bq == 'bc':
        Vxb = flavio.physics.ckm.get_ckm(par)[1,2]
    elif bq == 'bu':
        Vxb = flavio.physics.ckm.get_ckm(par)[0,2]
    # renormalization scale
    scale = flavio.config['renormalization scale']['bll']
    # Wilson coefficients
    wc = wc_obj.get_wc(bq + lep + 'nu' + nu, scale, par)
    # add SM contribution to Wilson coefficient
    if lep == nu:
        wc['CVL_'+bq+lep+'nu'+nu] += flavio.physics.bdecays.wilsoncoefficients.get_CVLSM(par, scale, nf=5)
    mb = flavio.physics.running.running.get_mb(par, scale)
    mu = 0  # neglecting up quark mass
    return br_plnu_general(wc, par, Vxb, B, bq, lep, nu, mb, mu, delta=0)

# function returning function needed for prediction instance
def br_blnu_fct(B, lep):
    def f(wc_obj, par):
        return br_blnu(wc_obj, par, B, lep)
    return f

# Observable and Prediction instances
_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_tex_B = {'B+': r'B^+', 'Bc': r'B_c'}
for l in ['e', 'mu', 'tau']:
    for B in ['B+', 'Bc']:
        _process_tex = _tex_B[B] + r"\to "+_tex[l]+r"^+\nu_"+_tex[l]
        _process_taxonomy = r'Process :: $b$ hadron decays :: Leptonic tree-level decays :: $B\to \ell\nu$ :: $' + _process_tex + r"$"

        _obs_name = "BR("+B+"->"+l+"nu)"
        _obs = flavio.classes.Observable(_obs_name)
        _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        flavio.classes.Prediction(_obs_name, br_blnu_fct(B, l))



r"""Functions for spectator scattering corrections to $B\to V\ell^+\ell^-$ decays.

This includes weak annihilation, chromomagnetic contributions, and light
quark-loop spectator scattering.
"""

import flavio
import numpy as np
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.physics.common import conjugate_par, conjugate_wc, add_dict
from flavio.config import config



# Auxiliary quantities and implementations

# function needed for the QCD factorization implementation (see qcdf.py)
def ha_qcdf_function(B, V):
    scale = config['renormalization scale']['bvll']
    label = meson_quark[(B,V)] + 'ee' # the lepton flavour is irrelevant here
                                      # as only dipole and 4-quark operators contribute!
    def function(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        wc = wctot_dict(wc_obj, label, scale, par)
        if cp_conjugate:
            wc = conjugate_wc(wc)
        return flavio.physics.bdecays.bvll.qcdf.helicity_amps_qcdf(q2, wc, par, B, V)
    return function

# ... and the same for the interpolated version (see qcdf_interpolate.py)
def ha_qcdf_interpolate_function(B, V, contribution='all'):
    scale = config['renormalization scale']['bvll']
    def function(wc_obj, par_dict, q2, cp_conjugate):
        return flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par_dict, B, V, cp_conjugate, contribution)
    return function

# loop over hadronic transitions and lepton flavours
# BTW, it is not necessary to loop over tau: for tautau final states, the minimum
# q2=4*mtau**2 is so high that QCDF is not valid anymore anyway!
for had in [('B0','K*0'), ('B+','K*+'), ('B0','rho0'), ('B+','rho+'), ('Bs','phi'), ]:
    process = had[0] + '->' + had[1] + 'll' # e.g. B0->K*0mumu
    quantity = process + ' spectator scattering'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2', 'cp_conjugate'])
    a.description = ('Contribution to ' + process + ' helicity amplitudes from'
                    ' non-factorizable spectator scattering.')

    # Implementation: QCD factorization
    iname = process + ' QCDF'
    i = Implementation(name=iname, quantity=quantity,
                   function=ha_qcdf_function(B=had[0], V=had[1]))
    i.set_description("QCD factorization")

    # Implementation: interpolated QCD factorization
    iname = process + ' QCDF interpolated'
    i = Implementation(name=iname, quantity=quantity,
                   function=ha_qcdf_interpolate_function(B=had[0], V=had[1]))
    i.set_description("Interpolated version of QCD factorization")



import unittest
import flavio
from flavio.classes import Observable, Prediction
from flavio.functions import get_dependent_parameters_sm

class TestFunctions(unittest.TestCase):
    def test_functions(self):
        o = Observable('test_obs')
        o.arguments = ['x']
        def f(wc_obj, par_dict, x):
            return x
        pr  = Prediction('test_obs', f )
        wc_obj = None
        self.assertEqual(flavio.sm_prediction('test_obs', 7), 7)
        self.assertEqual(flavio.np_prediction('test_obs', x=7, wc_obj=wc_obj), 7)
        self.assertEqual(flavio.sm_uncertainty('test_obs', 7), 0)
        self.assertEqual(flavio.np_uncertainty('test_obs', x=7, wc_obj=wc_obj), 0)
        self.assertEqual(flavio.sm_uncertainty('test_obs', 7, threads=2), 0)
        self.assertEqual(flavio.np_uncertainty('test_obs', x=7, wc_obj=wc_obj, threads=2), 0)
        # delete dummy instance
        Observable.del_instance('test_obs')

    def test_get_dep_par(self):
        self.assertEqual(
            get_dependent_parameters_sm('BR(Bs->mumu)'),
            {'DeltaGamma/Gamma_Bs', 'GF', 'Vcb', 'Vub', 'Vus', 'alpha_e', 'alpha_s', 'f_Bs', 'gamma', 'm_Bs', 'm_b', 'm_mu', 'm_s', 'tau_Bs', 'm_t'}
        )
        self.assertEqual(
            get_dependent_parameters_sm('BR(B0->ee)'),
            {'GF', 'Vcb', 'Vub', 'Vus', 'alpha_e', 'alpha_s', 'f_B0', 'gamma', 'm_B0', 'm_b', 'm_e', 'm_d', 'tau_B0', 'm_t'}
        )
        # for more complicated cases, just check there is no error
        get_dependent_parameters_sm('dBR/dq2(B+->Kmumu)', 3)
        get_dependent_parameters_sm('<dBR/dq2>(B+->Kmumu)', 3, 5)
        get_dependent_parameters_sm('dBR/dq2(B+->Kmumu)', q2=3)
        get_dependent_parameters_sm('<dBR/dq2>(B+->Kmumu)', q2min=3, q2max=5)



"""1D and 2D likelihood profilers for frequentist fits."""

import numpy as np
import flavio
import scipy.optimize
from flavio.math.optimize import minimize_robust, maximize_robust
from functools import partial
from multiprocessing import Pool
import warnings

def par_shift_scale(par_obj, parameters):
    """Determine a shift and scale factor that rescales
    parameters to be centered at 0 and varying by O(1)."""
    # central nuisance parameters
    cen_all = par_obj.get_central_all()
    # shift everything to 0
    shift = np.array([-cen_all[p] for p in parameters])
    # errors of nuisance parameters
    err_all = par_obj.get_1d_errors_rightleft()
    err = [err_all[p] for p in parameters]
    def scalefac(er, el):
        if er==0 and el==0:
            # for nuisances with vanishing uncertainties: don't rescale
            return 1
        # example: if errors are -0.01, +0.01, rescale by 10=1/0.1
        return 1/max(abs(er), abs(er))
    scale = np.array([scalefac(er, el) for er, el in err])
    return shift, scale

def nuisance_shift_scale(fit):
    """Determine a shift and scale factor that rescales nuisance
    parameters to be centered at 0 and varying by O(1)."""
    return par_shift_scale(fit.par_obj, fit.nuisance_parameters)

def reshuffle_1d(x, i0):
    """Reshuffle 1D array to make index i0 the first index and append the
    lower indices to the end."""
    return np.hstack((x[i0:], x[:i0]))

def unreshuffle_1d(x, i0):
    """Undo the reshuffle_1d operation."""
    N = len(x)
    return np.hstack((np.asarray(x)[N-i0:].T, np.asarray(x)[:N-i0].T)).T

def reshuffle_2d(x, ij0):
    """Reshuffle 2D array to make index (ij0) the first index, flatten the
    array, and append the following indices in the end."""
    m, n = np.asarray(x).shape
    x_rev = np.asarray(x)
    x_rev[1::2, :] = x_rev[1::2, ::-1] # reverse all odd rows
    i0, j0 = ij0
    if i0 % 2 != 0: # if row is odd
        j0 = n - j0 - 1 # flip the column index
    x_flat = x_rev.ravel()
    i0_1d = np.ravel_multi_index((i0, j0), (m, n))
    return reshuffle_1d(x_flat, i0_1d), i0_1d

def unreshuffle_2d(x, i0, shape):
    """Undo the reshuffle_2d operation."""
    x_flat = unreshuffle_1d(x, i0)
    x_rev = np.reshape(x_flat, shape)
    x_rev[1::2, :] = x_rev[1::2, ::-1] # reverse all odd rows
    return x_rev

def optimize_list_worker(x, n0, profiler, **kwargs):
    """Worker function needed for parallel execution of the likelihood
    optimization (see the `_optimize_list` method of `Profiler`)."""
    return profiler._optimize_list(x, n0, **kwargs)

class Profiler(object):
    """Parent class for profilers. Not meant to be used directly."""
    def __init__(self, fit):
        """Initialize the profiler instance."""
        self.fit = fit
        assert isinstance(fit, flavio.statistics.fits.FrequentistFit), \
                    "Fit object must be an instance of FrequentistFit"
        self.nuisance_shift, self.nuisance_scale = nuisance_shift_scale(fit)
        self.n_fit_p = len(self.fit.fit_parameters)
        self.n_nui_p = len(self.fit.nuisance_parameters)
        self.n_wc = len(self.fit.fit_wc_names)
        self.bf = None
        self.log_profile_likelihood = None
        self.profile_nuisance = None

    def f_target(self, x_n, par_wc_fixed):
        """Target function (log likelihood) in terms of rescaled and shifted
        nuisance parameters x_n for given values of the fixed parameters/Wilson
        coefficients."""
        x = np.zeros(self.fit.dimension)
        if self.n_fit_p > 0:
            x[:self.n_fit_p] = par_wc_fixed[:self.n_fit_p]
        if self.n_wc > 0:
            x[-self.n_wc:] = par_wc_fixed[-self.n_wc:]
        x[self.n_fit_p:self.n_fit_p+self.n_nui_p] = x_n/self.nuisance_scale - self.nuisance_shift
        if np.any(np.isnan(x)):
            return -np.inf
        try:
            return self.fit.log_likelihood(x)
        except ValueError:
            return -np.inf

    def f_target_global(self, x):
        """Target function (log likelihood) in terms of fit parameters,
        rescaled and shifted nuisance parameters, and (unscaled) Wilson
        coefficients for global optimization."""
        X = np.array(x)
        if self.n_nui_p > 0:
            X[self.n_fit_p:self.n_fit_p+self.n_nui_p] = x[self.n_fit_p:self.n_fit_p+self.n_nui_p]/self.nuisance_scale - self.nuisance_shift
        if np.any(np.isnan(x)):
            return -np.inf
        try:
            return self.fit.log_likelihood(X)
        except ValueError:
            return -np.inf

    def best_fit(self, fitpar0, **kwargs):
        """Determine the global best-fit point in the space of fit parameters,
        nuisance parameters, and Wilson coefficients.

        Returns a scipy.optimize.OptimizeResult instance."""
        x0 = np.zeros(self.fit.dimension)
        if self.n_fit_p > 0:
            x0[:self.n_fit_p] = fitpar0
        res = maximize_robust(self.f_target_global, x0=x0, **kwargs)
        return res

    def optimize_point(self, x, n0, **kwargs):
        """Maximize the nuisance likelihood for a single point x (a number
        for 1D profile likelihood and a tuple of two numbers for 2D), using
        n0 as initial values for the nuisance parameters.

        Returns z, n, n_scaled where
        - z is the optimized log-likelihood
        - n are the optimized nuisance parameters
        - n_scaled are the scaled and shifted optimized nuisance parameters
        """
        res = maximize_robust(self.f_target, args=(np.ravel([x]),), x0=n0, **kwargs)
        if res.success:
            z = res.fun
            n = res.x/self.nuisance_scale - self.nuisance_shift
            n_scaled = res.x
        else:
            z = np.nan
            n = np.nan
            n_scaled = np.nan
        return z, n, n_scaled

    def _optimize_list(self, x, n0, **kwargs):
        """Helper method for `optimize_list`."""
        z = np.zeros(len(x))
        n = np.zeros((len(x), self.n_nui_p))
        n0_i = n0
        for i, X in enumerate(x):
            z[i], n[i], n_scaled = self.optimize_point(x=X, n0=n0_i, **kwargs)
            if not np.any(np.isnan(n_scaled)):
                n0_i = n_scaled
        return z, n

    def optimize_list(self, x, n0, threads=1, **kwargs):
        """Maximize the nuisance likelihood for a list of points x, using
        n0 as initial values for the nuisance parameters at the first point.

        Returns z, n
        - z are the optimized log-likelihood values
        - n are the optimized nuisance parameters
        """
        if threads == 1:
            return self._optimize_list(x, n0, **kwargs)
        else:
            x_split = np.array_split(x, threads)
            with Pool(threads) as pool:
                zn = pool.map(partial(optimize_list_worker,
                              n0=n0,
                              profiler=self,
                              **kwargs),
                              x_split)
            z = np.concatenate([zi for zi, ni in zn])
            n = np.concatenate([ni for zi, ni in zn])
            return z, n


class Profiler1D(Profiler):
    """1-dimensional likelihood profiler.


    Methods:

    - run: profile the likelihood in the 1D interval.
    - pvalue_prob: return the p-value under the assumption of Wilks' theorem
    - pvalue_prob_plotdata: return a dictionary suited to be fed into the
      `flavio.plots.pvalue_plot` plotting function
    """
    def __init__(self, fit, x_min, x_max):
        """Initialize the profiler instance.

        Arguments:

        - x_min: lower bound of the parameter/coefficient of interest
        - x_max: upper bound of the parameter/coefficient of interest
        """
        assert len(fit.fit_parameters) + len(fit.fit_wc_names) == 1, (
                    "Fit instance must have precisely one fit parameter"
                    " or one fitted Wilson coefficient")
        super().__init__(fit)
        assert x_min < x_max, "x_max must be bigger than x_min"
        self.x_min = x_min
        self.x_max = x_max
        self.x_bf = None
        self.n_bf = None
        self.x = None

    def get_best_fit(self, **kwargs):
        """Determine the x-value of the best-fit point and save it."""
        try:
            bf = self.best_fit(fitpar0=self.fit.get_central_fit_parameters, **kwargs)
        except KeyError:
            # if the fit parameters do not have existing constraints, use center
            bf = self.best_fit(fitpar0=(self.x_max-self.x_min)/2, **kwargs)
        self.bf = bf
        if self.n_fit_p == 1:
            self.x_bf = bf.x[0] # best-fit parameter
            self.n_bf = bf.x[1:]
        elif self.n_wc == 1:
            self.x_bf = bf.x[-1] # ... or best-fit Wilson coefficient
            self.n_bf = bf.x[:-1]
        return self.x_bf

    def run(self, steps=20, threads=1, **kwargs):
        """Maximize the likelihood by varying the nuisance parameters.

        Arguments:

        - steps (defaults to 20): number of steps in the 1D interval of interest
        - threads (defaults to 1): number of parallel processes

        threads must be smaller than or equal to steps. Optimally, steps
        should be divisible by threads.

        Additional keyword arguments will be passed to
        `flavio.math.optimize.maximize_robust`.

        Returns:

        `x, z, n` where
        - x: the points at which the likelihood has been maximized
        - z: the log likelihood at these points
        - n: the values of the nuisance parameters at these points; has shape
            (n_nuisance, steps)
        """
        if threads > steps:
            raise ValueError("Number of threads cannot be larger than number of steps!")
        # determine the x-value of the best-fit point
        self.get_best_fit(**kwargs)
        if self.x_bf <= self.x_min or self.x_bf >= self.x_max:
            # if the best-fit value is at the border or outside,
            # just make a linspace
            x = np.linspace(self.x_min, self.x_max, steps)
        else:
            # otherwise, divide the range into x<x_bf and x>x_bf,
            # with the same number of steps on both sides
            steps_l = steps//2
            steps_r = steps - steps_l
            x = np.hstack([np.linspace(self.x_min, self.x_bf, steps_l),
                           np.linspace(self.x_bf, self.x_max, steps_r)])
        # determine index in x-array where the x is closest to x_bf
        i0 = (np.abs(x-self.x_bf)).argmin()
        x = reshuffle_1d(x, i0)
        # optimize, starting with global best-fit values for nuisance parameters
        z, n = self.optimize_list(x=x, n0=self.n_bf, threads=threads, **kwargs)
        x = unreshuffle_1d(x, i0)
        z = unreshuffle_1d(z, i0)
        n = unreshuffle_1d(n, i0).T
        self.x = x
        self.log_profile_likelihood = z
        self.profile_nuisance = n
        return x, z, n

    def pvalue_prob(self):
        """Return p-value obtained under the assumption of Wilks' theorem.

        Requires the profiler to be run first using the `run` method."""
        if self.log_profile_likelihood is None:
            warnings.warn("You must run the profiler first.")
            return None
        chi2_dist = scipy.stats.chi2(1)
        chi2 = -2*self.log_profile_likelihood
        chi2_bf = -2*self.bf.fun
        # normally, the minimal chi2 should of course be the chi2 at the best-fit
        # point. If for reasons of numerical inaccuracy the chi2 array contains
        # a smaller value, use that.
        chi2_min = min(np.min(chi2), chi2_bf)
        delta_chi2 = chi2 - chi2_min
        cl = chi2_dist.cdf(delta_chi2)
        return 1-cl

    def pvalue_prob_plotdata(self):
        """Return a dictionary that can be fed into `flavio.plots.pvalue_plot`."""
        return {
                'x': self.x,
                'y': self.pvalue_prob(),
               }


class Profiler2D(Profiler):
    """2-dimensional likelihood profiler.


    Methods:

    - run: profile the likelihood in the 2D plane.
    - contour_plotdata: return a dictionary suited to be fed into the
      `flavio.plots.contour` plotting function
    """
    def __init__(self, fit, x_min, x_max, y_min, y_max):
        """Initialize the profiler instance.

        Arguments:

        - x_min, x_max: lower and upper bounds of x
        - y_min, y_max: lower and upper bounds of y
        """
        assert len(fit.fit_parameters) + len(fit.fit_wc_names) == 2, (
                    "Fit instance must have precisely one fit parameter"
                    " or one fitted Wilson coefficient")
        super().__init__(fit)
        assert x_min < x_max, "x_max must be bigger than x_min"
        assert y_min < y_max, "y_max must be bigger than y_min"
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.x_bf = None
        self.y_bf = None

    def get_best_fit(self, **kwargs):
        """Determine the position of the best-fit point and save it."""
        try:
            bf = self.best_fit(fitpar0=self.fit.get_central_fit_parameters, **kwargs)
        except KeyError:
            bf = self.best_fit(fitpar0=[(self.x_max-self.x_min)/2,
                                        (self.y_max-self.y_min)/2], **kwargs)
        self.bf = bf
        if self.n_wc > 0:
            self.x_bf, self.y_bf = np.hstack((bf.x[:self.n_fit_p], bf.x[-self.n_wc:]))
        else:
            self.x_bf, self.y_bf = bf.x[:self.n_fit_p]

    def run(self, steps=(10, 10), usebf=False, threads=1, **kwargs):

        """Maximize the likelihood by varying the nuisance parameters.

        Arguments:

        - steps: number of steps in the in the x and y direction.
          Tuple of length 2 that defaults to (10, 10)
        - threads (defaults to 1): number of parallel processes
        - method: minimization method to be used by scipy.optimize.minimize

        threads must be smaller than or equal to the product of steps in x and
        y direction (i.e., the number of grid points). Optimally, the number of
        grid points should be a multiple of the number of threads.

        Returns:

        `x, y, z, n` where
        - x, y: the points at which the likelihood has been maximized
        - z: the log likelihood at these points
        - n: the values of the nuisance parameters at these points; has shape
          (n_nuisance, steps_x, steps_y)
        """
        if threads > steps[0]*steps[1]:
            raise ValueError("Number of threads cannot be larger than number of grid points!")
        # determine x- and y-value at the global best-fit point
        if usebf:
            self.get_best_fit(**kwargs)
        x = np.linspace(self.x_min, self.x_max, steps[0])
        y = np.linspace(self.y_min, self.y_max, steps[1])
        if usebf:
            # determine index in x and y-arrays where the x/y is closest to x_bf/y_bf
            ij0 = (np.abs(x-self.x_bf)).argmin(), (np.abs(x-self.x_bf)).argmin()
        else:
            # else, just use the center
            ij0 = (steps[0]//2, steps[1]//2)
        z = np.zeros(steps)
        n = np.zeros((self.n_nui_p, steps[0], steps[1]))
        xx, yy = np.meshgrid(x, y, indexing='ij')
        xx, i0_1d = reshuffle_2d(xx, ij0)
        yy, i0_1d = reshuffle_2d(yy, ij0)
        z, i0_1d = reshuffle_2d(z, ij0)
        n = np.array([reshuffle_2d(ni, ij0)[0] for ni in n])
        # start with global best-fit values for nuisance parameters
        if usebf:
            n0 = self.bf.x[self.n_fit_p:self.n_fit_p+self.n_nui_p]
        else:
            n0 = self.fit.get_central_nuisance_parameters
        z, n = self.optimize_list(x=np.transpose([xx, yy]),
                                  n0=n0,
                                  threads=threads,
                                  **kwargs)
        xx = unreshuffle_2d(xx, i0_1d, steps)
        yy = unreshuffle_2d(yy, i0_1d, steps)
        x = xx[:,0]
        y = yy[0]
        z = unreshuffle_2d(z, i0_1d, steps)
        n = np.array([unreshuffle_2d(ni, i0_1d, steps) for ni in n.T])
        self.x = x
        self.y = y
        self.log_profile_likelihood = z
        self.profile_nuisance = n
        return x, y, z, n

    def contour_plotdata(self, n_sigma=(1,2)):
        """Return a dictionary that can be fed into `flavio.plots.contour`.

        Parameters:

        - `n_sigma`: tuple with integer sigma values which should be plotted.
          Defaults to (1, 2).
        """
        deltachi2 = -2*(self.log_profile_likelihood
                        -np.nanmax(self.log_profile_likelihood))
        x, y = np.meshgrid(self.x, self.y, indexing='ij')
        return {
                'x': x,
                'y': y,
                'z': deltachi2,
                'levels': tuple(flavio.statistics.functions.delta_chi2(n, 2)
                                for n in n_sigma),
               }



r"""Functions to parametrize subleading hadronic effects in $B\to V\ell^+\ell^-$
decays."""


import flavio
import numpy as np
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.physics.common import conjugate_par, conjugate_wc, add_dict
from flavio.config import config


# First, we define functions that allow us to get a contribution to the
# amplitudes from an effective, helicity dependent shift in either C7 or C9

# auxiliary function to construct helicity_amps_deltaC7, helicity_amps_deltaC9
def _helicity_amps_deltaC(q2, deltaC, C_name, par, B, V):
    mB = par['m_'+B]
    mV = par['m_'+V]
    scale = config['renormalization scale']['bvll']
    mb = flavio.physics.running.running.get_mb(par, scale)
    N = flavio.physics.bdecays.bvll.amplitudes.prefactor(q2, par, B, V)
    ff = flavio.physics.bdecays.bvll.amplitudes.get_ff(q2, par, B, V)
    wc  = {'7': 0, '7p': 0, 'v': 0, 'a': 0, 's': 0, 'p': 0, 't': 0,'vp': 0, 'ap': 0, 'sp': 0, 'pp': 0, 'tp': 0, }
    wc[C_name] = deltaC
    return flavio.physics.bdecays.angular.helicity_amps_v(q2, mB, mV, mb, 0, 0, 0, ff, wc, N) # ml=0 as this only affects scalar contributions

def helicity_amps_deltaC7(q2, deltaC7_dict, par, B, V):
    r"""A function returning a contribution to the helicity amplitudes in
    $B\to V\ell^+\ell^-$ coming from an effective helicity-dependent shift of
    the Wilson coefficient $C_7(\mu_b)$. This can be used to parametrize
    residual uncertainties due to subleading non-factorizable hadronic effects.

    The input dictionary `deltaC7_dict` should be of the form

    `{ ('0','V'): deltaC7_0, ('pl','V'): deltaC7_pl, ('mi','V'): deltaC7_mi }`
    """
    ha = {}
    for amp in [('0','V'), ('pl','V'), ('mi','V')]:
        ha[amp] = _helicity_amps_deltaC(q2, deltaC7_dict[amp], '7', par, B, V)[amp]
    return ha

def helicity_amps_deltaC7p(q2, deltaC7p_dict, par, B, V):
    r"""A function returning a contribution to the helicity amplitudes in
    $B\to V\ell^+\ell^-$ coming from an effective helicity-dependent shift of
    the Wilson coefficient $C_7'(\mu_b)$. This can be used to parametrize
    residual uncertainties due to subleading non-factorizable hadronic effects.

    The input dictionary `deltaC7p_dict` should be of the form

    `{ ('0','V'): deltaC7p_0, ('pl','V'): deltaC7p_pl, ('mi','V'): deltaC7p_mi }`
    """
    ha = {}
    for amp in [('0','V'), ('pl','V'), ('mi','V')]:
        ha[amp] = _helicity_amps_deltaC(q2, deltaC7p_dict[amp], '7p', par, B, V)[amp]
    return ha

def helicity_amps_deltaC9(q2, deltaC9_dict, par, B, V):
    r"""A function returning a contribution to the helicity amplitudes in
    $B\to V\ell^+\ell^-$ coming from an effective helicity-dependent shift of
    the Wilson coefficient $C_7(\mu_b)$. This can be used to parametrize
    residual uncertainties due to subleading non-factorizable hadronic effects.

    The input dictionary `deltaC9_dict` should be of the form

    `{ ('0','V'): deltaC9_0, ('pl','V'): deltaC9_pl, ('mi','V'): deltaC9_mi }`
    """
    ha = {}
    for amp in [('0','V'), ('pl','V'), ('mi','V')]:
        ha[amp] = _helicity_amps_deltaC(q2, deltaC9_dict[amp], 'v', par, B, V)[amp]
    return ha

# One possibility is to parametrize the effective shift in C7 or C9 as a simple
# polynomial in q2.

def helicity_amps_deltaC7_polynomial(q2, par, B, V):
    deltaC7_0   =( par[B+'->'+V+' deltaC7 a_0 Re']  + par[B+'->'+V+' deltaC7 b_0 Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_0 Im']  + par[B+'->'+V+' deltaC7 b_0 Im'] *q2 ))
    deltaC7_pl  =( par[B+'->'+V+' deltaC7 a_+ Re']  + par[B+'->'+V+' deltaC7 b_+ Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_+ Im']  + par[B+'->'+V+' deltaC7 b_+ Im'] *q2 ))
    deltaC7_mi  =( par[B+'->'+V+' deltaC7 a_- Re']  + par[B+'->'+V+' deltaC7 b_- Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_- Im']  + par[B+'->'+V+' deltaC7 b_- Im'] *q2 ))
    deltaC7_dict = { ('0','V'): deltaC7_0, ('pl','V'): deltaC7_pl, ('mi','V'): deltaC7_mi }
    return helicity_amps_deltaC7(q2, deltaC7_dict, par, B, V)

def helicity_amps_deltaC7C7p_polynomial(q2, par, B, V):
    deltaC7_0   =( par[B+'->'+V+' deltaC7 a_0 Re']  + par[B+'->'+V+' deltaC7 b_0 Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_0 Im']  + par[B+'->'+V+' deltaC7 b_0 Im'] *q2 ))
    deltaC7p_pl  =( par[B+'->'+V+' deltaC7p a_+ Re']  + par[B+'->'+V+' deltaC7p b_+ Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7p a_+ Im']  + par[B+'->'+V+' deltaC7p b_+ Im'] *q2 ))
    deltaC7_mi  =( par[B+'->'+V+' deltaC7 a_- Re']  + par[B+'->'+V+' deltaC7 b_- Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_- Im']  + par[B+'->'+V+' deltaC7 b_- Im'] *q2 ))
    deltaC7_dict = { ('0','V'): deltaC7_0, ('pl','V'): 0, ('mi','V'): deltaC7_mi }
    deltaC7p_dict = { ('0','V'): 0, ('pl','V'): deltaC7p_pl, ('mi','V'): 0 }
    ha_deltaC7 = helicity_amps_deltaC7(q2, deltaC7_dict, par, B, V)
    ha_deltaC7p = helicity_amps_deltaC7p(q2, deltaC7p_dict, par, B, V)
    return add_dict([ha_deltaC7, ha_deltaC7p])

def helicity_amps_deltaC9C7C7p_polynomial(q2, par, B, V):
    deltaC9_0   =( par[B+'->'+V+' deltaC9 a_0 Re']  + par[B+'->'+V+' deltaC9 b_0 Re'] *q2
             +1j*( par[B+'->'+V+' deltaC9 a_0 Im']  + par[B+'->'+V+' deltaC9 b_0 Im'] *q2 ))
    deltaC7p_pl  =( par[B+'->'+V+' deltaC7p a_+ Re']  + par[B+'->'+V+' deltaC7p b_+ Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7p a_+ Im']  + par[B+'->'+V+' deltaC7p b_+ Im'] *q2 ))
    deltaC7_mi  =( par[B+'->'+V+' deltaC7 a_- Re']  + par[B+'->'+V+' deltaC7 b_- Re'] *q2
             +1j*( par[B+'->'+V+' deltaC7 a_- Im']  + par[B+'->'+V+' deltaC7 b_- Im'] *q2 ))
    deltaC9_dict = { ('0','V'): deltaC9_0, ('pl','V'): 0, ('mi','V'): 0 }
    deltaC7_dict = { ('0','V'): 0, ('pl','V'): 0, ('mi','V'): deltaC7_mi }
    deltaC7p_dict = { ('0','V'): 0, ('pl','V'): deltaC7p_pl, ('mi','V'): 0 }
    ha_deltaC9 = helicity_amps_deltaC7(q2, deltaC9_dict, par, B, V)
    ha_deltaC7 = helicity_amps_deltaC7(q2, deltaC7_dict, par, B, V)
    ha_deltaC7p = helicity_amps_deltaC7p(q2, deltaC7p_dict, par, B, V)
    return add_dict([ha_deltaC7, ha_deltaC7p])

# note that, when parametrizing the correction as a shift to C9 rather than C7,
# the contribution to the transverse (+ and -) amplitudes has to start with
# 1/q2, otherwise one effectively sets corrections to B->Vgamma to zero.
def helicity_amps_deltaC9_polynomial(q2, par, B, V):
    deltaC9_0   =( par[B+'->'+V+' deltaC9 a_0 Re']  + par[B+'->'+V+' deltaC9 b_0 Re'] *q2
             +1j*( par[B+'->'+V+' deltaC9 a_0 Im']  + par[B+'->'+V+' deltaC9 b_0 Im'] *q2 ))
    deltaC9_pl  =( par[B+'->'+V+' deltaC9 a_+ Re']/q2  + par[B+'->'+V+' deltaC9 b_+ Re']
             +1j*( par[B+'->'+V+' deltaC9 a_+ Im']/q2  + par[B+'->'+V+' deltaC9 b_+ Im'] ))
    deltaC9_mi  =( par[B+'->'+V+' deltaC9 a_- Re']/q2  + par[B+'->'+V+' deltaC9 b_- Re']
             +1j*( par[B+'->'+V+' deltaC9 a_- Im']/q2  + par[B+'->'+V+' deltaC9 b_- Im'] ))
    deltaC9_dict = { ('0','V'): deltaC9_0, ('pl','V'): deltaC9_pl, ('mi','V'): deltaC9_mi }
    return helicity_amps_deltaC9(q2, deltaC9_dict, par, B, V)

# a constant shift, e.g. for high q^2
def helicity_amps_deltaC9_constant(q2, par, B, V):
    deltaC9_0   = par[B+'->'+V+' deltaC9 c_0 Re'] + 1j*par[B+'->'+V+' deltaC9 c_0 Im']
    deltaC9_pl   = par[B+'->'+V+' deltaC9 c_+ Re'] + 1j*par[B+'->'+V+' deltaC9 c_+ Im']
    deltaC9_mi   = par[B+'->'+V+' deltaC9 c_- Re'] + 1j*par[B+'->'+V+' deltaC9 c_- Im']
    deltaC9_dict = { ('0','V'): deltaC9_0, ('pl','V'): deltaC9_pl, ('mi','V'): deltaC9_mi }
    return helicity_amps_deltaC9(q2, deltaC9_dict, par, B, V)


# Functions returning functions needed for Implementation
def fct_deltaC7_polynomial(B, V):
    def fct(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        return helicity_amps_deltaC7_polynomial(q2, par, B, V)
    return fct

def fct_deltaC7C7p_polynomial(B, V):
    def fct(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        return helicity_amps_deltaC7C7p_polynomial(q2, par, B, V)
    return fct

def fct_deltaC9C7C7p_polynomial(B, V):
    def fct(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        return helicity_amps_deltaC9C7C7p_polynomial(q2, par, B, V)
    return fct

def fct_deltaC9_polynomial(B, V):
    def fct(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        return helicity_amps_deltaC9_polynomial(q2, par_dict, B, V)
    return fct

def fct_deltaC9_constant(B, V):
    def fct(wc_obj, par_dict, q2, cp_conjugate):
        par = par_dict.copy()
        if cp_conjugate:
            par = conjugate_par(par)
        return helicity_amps_deltaC9_constant(q2, par_dict, B, V)
    return fct

# AuxiliaryQuantity & Implementation: subleading effects at LOW q^2

for had in [('B0','K*0'), ('B+','K*+'), ('Bs','phi'), ]:
    process = had[0] + '->' + had[1] + 'll' # e.g. B0->K*0mumu
    quantity = process + ' subleading effects at low q2'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2', 'cp_conjugate'])
    a.description = ('Contribution to ' + process + ' helicity amplitudes from'
                    ' subleading hadronic effects (i.e. all effects not included'
                    r'elsewhere) at $q^2$ below the charmonium resonances')

    # Implementation: C7-polynomial
    iname = process + ' deltaC7 polynomial'
    i = Implementation(name=iname, quantity=quantity,
                   function=fct_deltaC7_polynomial(B=had[0], V=had[1]))
    i.set_description(r"Effective shift in the Wilson coefficient $C_7(\mu_b)$"
                      r" as a first-order polynomial in $q^2$.")

    # Implementation: C7-C7'-polynomial
    iname = process + ' deltaC7, 7p polynomial'
    i = Implementation(name=iname, quantity=quantity,
                   function=fct_deltaC7C7p_polynomial(B=had[0], V=had[1]))
    i.set_description(r"Effective shift in the Wilson coefficient $C_7(\mu_b)$"
                      r" (in the $0$ and $-$ helicity amplitudes) and"
                      r" $C_7'(\mu_b)$ (in the $+$ helicity amplitude)"
                      r" as a first-order polynomial in $q^2$.")

    # Implementation: C9-C7-C7'-polynomial
    iname = process + ' deltaC9, 7, 7p polynomial'
    i = Implementation(name=iname, quantity=quantity,
                   function=fct_deltaC9C7C7p_polynomial(B=had[0], V=had[1]))
    i.set_description(r"Effective shift in the Wilson coefficient $C_9(\mu_b)$"
                      r" (in the $0$ helicity amplitude), $C_7(\mu_b)$"
                      r" (in the $-$ helicity amplitude) and"
                      r" $C_7'(\mu_b)$ (in the $+$ helicity amplitude)"
                      r" as a first-order polynomial in $q^2$.")

    # Implementation: C9-polynomial
    iname = process + ' deltaC9 polynomial'
    i = Implementation(name=iname, quantity=quantity,
                   function=fct_deltaC9_polynomial(B=had[0], V=had[1]))
    i.set_description(r"Effective shift in the Wilson coefficient $C_9(\mu_b)$"
                      r" as a first-order polynomial in $q^2$.")


# AuxiliaryQuantity & Implementation: subleading effects at HIGH q^2

for had in [('B0','K*0'), ('B+','K*+'), ('Bs','phi'), ]:
    process = had[0] + '->' + had[1] + 'll' # e.g. B0->K*0mumu
    quantity = process + ' subleading effects at high q2'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2', 'cp_conjugate'])
    a.description = ('Contribution to ' + process + ' helicity amplitudes from'
                    ' subleading hadronic effects (i.e. all effects not included'
                    r'elsewhere) at $q^2$ above the charmonium resonances')

    # Implementation: C9 constant shift
    iname = process + ' deltaC9 shift'
    i = Implementation(name=iname, quantity=quantity,
                   function=fct_deltaC9_constant(B=had[0], V=had[1]))
    i.set_description(r"Effective constant shift in the Wilson coefficient $C_9(\mu_b)$.")



import unittest
import numpy as np
import flavio

wc_sm = flavio.physics.eft.WilsonCoefficients()
wc_np = flavio.physics.eft.WilsonCoefficients()
wc_np.set_initial({'C10_bsemu':4., 'C10_bsmue':2.}, 160.)

class TestLFV(unittest.TestCase):
    def test_lfv(self):
        obs_1 = flavio.classes.Observable["BR(B0->K*emu)"]
        obs_2 = flavio.classes.Observable["BR(B0->K*mue)"]
        self.assertEqual(obs_1.prediction_central(flavio.default_parameters, wc_sm), 0)
        # BR(B->K*emu) should be 4 times larger as Wilson coeff is 2x the mue one
        self.assertAlmostEqual(
            obs_1.prediction_central(flavio.default_parameters, wc_np)
            /obs_2.prediction_central(flavio.default_parameters, wc_np),
            4.,  places=10)
        # test for errors
        flavio.sm_prediction("BR(B+->K*mue)")
        flavio.sm_prediction("BR(B0->rhotaue)")
        flavio.sm_prediction("BR(B+->rhotaumu)")
        flavio.sm_prediction("BR(Bs->phimutau)")



from setuptools import setup, find_packages

with open('flavio/_version.py') as f:
    exec(f.read())

setup(name='flavio',
      version=__version__,
      author='David M. Straub',
      author_email='david.straub@tum.de',
      url='https://flav-io.github.io',
      description='A Python package for flavour physics phenomenology in the Standard Model and beyond',
      long_description="""``flavio`` is a package to compute observables in flavour physics
      both within the Standard Model of particle physics and in the presence of new
      physics encoded in Wilson coefficients of local dimension-6 operators.
      Observables implemented include rare B meson decays and meson-antimeson
      mixing.""",
      license='MIT',
      packages=find_packages(),
      package_data={
      'flavio':['data/*.yml',
                'data/test/*',
                'physics/data/arXiv-0810-4077v3/*',
                'physics/data/arXiv-1503-05534v1/*',
                'physics/data/arXiv-1503-05534v2/*',
                'physics/data/arXiv-1501-00367v2/*',
                'physics/data/arXiv-1602-01399v1/*',
                'physics/data/pdg/*',
                'physics/data/qcdf_interpolate/*',
                'physics/data/wcsm/*',
                ]
      },
      install_requires=['numpy', 'scipy>=0.14', 'setuptools>=3.3', 'pyyaml', 'mpmath',
                        'wcxf>=1.2.1', 'ckmutil', 'wetrunner'],
      extras_require={
            'testing': ['nose'],
            'plotting': ['matplotlib>=1.4'],
            'sampling': ['pypmc>=1.1', 'emcee', 'iminuit',],
            },
    )
